@@ -35,8 +35,8 @@ static const struct compopt compopts[] = {
 	{"--output-directory", AFFECTS_CPP | TAKES_ARG}, // nvcc
 	{"--param",         TAKES_ARG},
 	{"--save-temps",    TOO_HARD},
-	{"--save-temps=cwd",TOO_HARD},
-	{"--save-temps=obj",TOO_HARD},
+	{"--save-temps=cwd", TOO_HARD},
+	{"--save-temps=obj", TOO_HARD},
 	{"--serialize-diagnostics", TAKES_ARG | TAKES_PATH},
 	{"-A",              TAKES_ARG},
 	{"-B",              TAKES_ARG | TAKES_CONCAT_ARG | TAKES_PATH},
@@ -142,6 +142,9 @@ compopt_short(bool (*fn)(const char *), const char *option)
 	return retval;
 }
 
+// Used by unittest/test_compopt.c.
+bool compopt_verify_sortedness(void);
+
 // For test purposes.
 bool
 compopt_verify_sortedness(void)
