@@ -19,6 +19,7 @@
 
 typedef bool (*conf_item_parser)(const char *str, void *result, char **errmsg);
 typedef bool (*conf_item_verifier)(void *value, char **errmsg);
+typedef const char *(*conf_item_formatter)(void *value);
 
 struct conf_item {
 	const char *name;
@@ -26,6 +27,7 @@ struct conf_item {
 	conf_item_parser parser;
 	size_t offset;
 	conf_item_verifier verifier;
+	conf_item_formatter formatter;
 };
 
 struct env_to_conf_item {
@@ -50,6 +52,19 @@ parse_bool(const char *str, void *result, char **errmsg)
 	}
 }
 
+static const char *
+bool_to_string(bool value)
+{
+	return value ? "true" : "false";
+}
+
+static const char *
+format_bool(void *value)
+{
+	bool *b = (bool *)value;
+	return x_strdup(bool_to_string(*b));
+}
+
 static bool
 parse_env_string(const char *str, void *result, char **errmsg)
 {
@@ -59,6 +74,19 @@ parse_env_string(const char *str, void *result, char **errmsg)
 	return *value != NULL;
 }
 
+static const char *
+format_string(void *value)
+{
+	char **str = (char **)value;
+	return x_strdup(*str);
+}
+
+static const char *
+format_env_string(void *value)
+{
+	return format_string(value);
+}
+
 static bool
 parse_float(const char *str, void *result, char **errmsg)
 {
@@ -75,6 +103,13 @@ parse_float(const char *str, void *result, char **errmsg)
 	}
 }
 
+static const char *
+format_float(void *value)
+{
+	float *x = (float *)value;
+	return format("%.1f", *x);
+}
+
 static bool
 parse_size(const char *str, void *result, char **errmsg)
 {
@@ -89,6 +124,13 @@ parse_size(const char *str, void *result, char **errmsg)
 	}
 }
 
+static const char *
+format_size(void *value)
+{
+	uint64_t *size = (uint64_t *)value;
+	return format_parsable_size_with_suffix(*size);
+}
+
 static bool
 parse_sloppiness(const char *str, void *result, char **errmsg)
 {
@@ -106,6 +148,8 @@ parse_sloppiness(const char *str, void *result, char **errmsg)
 			*value |= SLOPPY_FILE_MACRO;
 		} else if (str_eq(word, "file_stat_matches")) {
 			*value |= SLOPPY_FILE_STAT_MATCHES;
+		} else if (str_eq(word, "file_stat_matches_ctime")) {
+			*value |= SLOPPY_FILE_STAT_MATCHES_CTIME;
 		} else if (str_eq(word, "include_file_ctime")) {
 			*value |= SLOPPY_INCLUDE_FILE_CTIME;
 		} else if (str_eq(word, "include_file_mtime")) {
@@ -127,6 +171,42 @@ parse_sloppiness(const char *str, void *result, char **errmsg)
 	return true;
 }
 
+static const char *
+format_sloppiness(void *value)
+{
+	unsigned *sloppiness = (unsigned *)value;
+	char *s = x_strdup("");
+	if (*sloppiness & SLOPPY_FILE_MACRO) {
+		reformat(&s, "%sfile_macro, ", s);
+	}
+	if (*sloppiness & SLOPPY_INCLUDE_FILE_MTIME) {
+		reformat(&s, "%sinclude_file_mtime, ", s);
+	}
+	if (*sloppiness & SLOPPY_INCLUDE_FILE_CTIME) {
+		reformat(&s, "%sinclude_file_ctime, ", s);
+	}
+	if (*sloppiness & SLOPPY_TIME_MACROS) {
+		reformat(&s, "%stime_macros, ", s);
+	}
+	if (*sloppiness & SLOPPY_PCH_DEFINES) {
+		reformat(&s, "%spch_defines, ", s);
+	}
+	if (*sloppiness & SLOPPY_FILE_STAT_MATCHES) {
+		reformat(&s, "%sfile_stat_matches, ", s);
+	}
+	if (*sloppiness & SLOPPY_FILE_STAT_MATCHES_CTIME) {
+		reformat(&s, "%sfile_stat_matches_ctime, ", s);
+	}
+	if (*sloppiness & SLOPPY_NO_SYSTEM_HEADERS) {
+		reformat(&s, "%sno_system_headers, ", s);
+	}
+	if (*sloppiness) {
+		// Strip last ", ".
+		s[strlen(s) - 2] = '\0';
+	}
+	return s;
+}
+
 static bool
 parse_string(const char *str, void *result, char **errmsg)
 {
@@ -158,6 +238,17 @@ parse_umask(const char *str, void *result, char **errmsg)
 	}
 }
 
+static const char *
+format_umask(void *value)
+{
+	unsigned *umask = (unsigned *)value;
+	if (*umask == UINT_MAX) {
+		return x_strdup("");
+	} else {
+		return format("%03o", *umask);
+	}
+}
+
 static bool
 parse_unsigned(const char *str, void *result, char **errmsg)
 {
@@ -175,9 +266,10 @@ parse_unsigned(const char *str, void *result, char **errmsg)
 }
 
 static const char *
-bool_to_string(bool value)
+format_unsigned(void *value)
 {
-	return value ? "true" : "false";
+	unsigned *i = (unsigned *)value;
+	return format("%u", *i);
 }
 
 static bool
@@ -210,9 +302,10 @@ verify_dir_levels(void *value, char **errmsg)
 }
 
 #define ITEM(name, type) \
-  parse_ ## type, offsetof(struct conf, name), NULL
+	parse_ ## type, offsetof(struct conf, name), NULL, format_ ## type
 #define ITEM_V(name, type, verification) \
-  parse_ ## type, offsetof(struct conf, name), verify_ ## verification
+	parse_ ## type, offsetof(struct conf, name), \
+	verify_ ## verification, format_ ## type
 
 #include "confitems_lookup.c"
 #include "envtoconfitems_lookup.c"
@@ -253,8 +346,8 @@ handle_conf_setting(struct conf *conf, const char *key, const char *value,
 			fatal("invalid boolean environment variable value \"%s\"", value);
 		}
 
-		bool *value = (bool *)((char *)conf + item->offset);
-		*value = !negate_boolean;
+		bool *boolvalue = (bool *)((char *)conf + item->offset);
+		*boolvalue = !negate_boolean;
 		goto out;
 	}
 
@@ -328,6 +421,7 @@ conf_create(void)
 	conf->compression = false;
 	conf->compression_level = 6;
 	conf->cpp_extension = x_strdup("");
+	conf->debug = false;
 	conf->direct_mode = true;
 	conf->disable = false;
 	conf->extra_files_to_hash = x_strdup("");
@@ -340,6 +434,7 @@ conf_create(void)
 	conf->max_files = 0;
 	conf->max_size = (uint64_t)5 * 1000 * 1000 * 1000;
 	conf->path = x_strdup("");
+	conf->pch_external_checksum = false;
 	conf->prefix_command = x_strdup("");
 	conf->prefix_command_cpp = x_strdup("");
 	conf->read_only = false;
@@ -376,7 +471,7 @@ conf_free(struct conf *conf)
 	free(conf->prefix_command);
 	free(conf->prefix_command_cpp);
 	free(conf->temporary_dir);
-	free(conf->item_origins);
+	free((void *)conf->item_origins); // Workaround for MSVC warning
 	free(conf);
 }
 
@@ -461,9 +556,10 @@ conf_update_from_environment(struct conf *conf, char **errmsg)
 		}
 
 		char *errmsg2;
-		if (!handle_conf_setting(
-		      conf, env_to_conf_item->conf_name, q, &errmsg2, true, negate,
-		      "environment")) {
+		bool ok = handle_conf_setting(
+			conf, env_to_conf_item->conf_name, q, &errmsg2, true, negate,
+			"environment");
+		if (!ok) {
 			*errmsg = format("%s: %s", key, errmsg2);
 			free(errmsg2);
 			free(key);
@@ -533,148 +629,81 @@ conf_set_value_in_file(const char *path, const char *key, const char *value,
 	return true;
 }
 
+bool
+conf_print_value(struct conf *conf, const char *key,
+                 FILE *file, char **errmsg)
+{
+	const struct conf_item *item = find_conf(key);
+	if (!item) {
+		*errmsg = format("unknown configuration option \"%s\"", key);
+		return false;
+	}
+	void *value = (char *)conf + item->offset;
+	char *str = (char *)item->formatter(value);
+	fprintf(file, "%s\n", str);
+	free(str);
+	return true;
+}
+
+static bool
+print_item(struct conf *conf, const char *key,
+           void (*printer)(const char *descr, const char *origin,
+                           void *context),
+           void *context)
+{
+	const struct conf_item *item = find_conf(key);
+	if (!item) {
+		return false;
+	}
+	void *value = (char *)conf + item->offset;
+	char *str = (char *)item->formatter(value);
+	char *buf = x_strdup("");
+	reformat(&buf, "%s = %s", key, str);
+	printer(buf, conf->item_origins[item->number], context);
+	free(buf);
+	free(str);
+	return true;
+}
+
 bool
 conf_print_items(struct conf *conf,
                  void (*printer)(const char *descr, const char *origin,
                                  void *context),
                  void *context)
 {
-	char *s = x_strdup("");
-
-	reformat(&s, "base_dir = %s", conf->base_dir);
-	printer(s, conf->item_origins[find_conf("base_dir")->number], context);
-
-	reformat(&s, "cache_dir = %s", conf->cache_dir);
-	printer(s, conf->item_origins[find_conf("cache_dir")->number], context);
-
-	reformat(&s, "cache_dir_levels = %u", conf->cache_dir_levels);
-	printer(s, conf->item_origins[find_conf("cache_dir_levels")->number],
-	        context);
-
-	reformat(&s, "compiler = %s", conf->compiler);
-	printer(s, conf->item_origins[find_conf("compiler")->number], context);
-
-	reformat(&s, "compiler_check = %s", conf->compiler_check);
-	printer(s, conf->item_origins[find_conf("compiler_check")->number], context);
-
-	reformat(&s, "compression = %s", bool_to_string(conf->compression));
-	printer(s, conf->item_origins[find_conf("compression")->number], context);
-
-	reformat(&s, "compression_level = %u", conf->compression_level);
-	printer(s, conf->item_origins[find_conf("compression_level")->number],
-	        context);
-
-	reformat(&s, "cpp_extension = %s", conf->cpp_extension);
-	printer(s, conf->item_origins[find_conf("cpp_extension")->number], context);
-
-	reformat(&s, "direct_mode = %s", bool_to_string(conf->direct_mode));
-	printer(s, conf->item_origins[find_conf("direct_mode")->number], context);
-
-	reformat(&s, "disable = %s", bool_to_string(conf->disable));
-	printer(s, conf->item_origins[find_conf("disable")->number], context);
-
-	reformat(&s, "extra_files_to_hash = %s", conf->extra_files_to_hash);
-	printer(s, conf->item_origins[find_conf("extra_files_to_hash")->number],
-	        context);
-
-	reformat(&s, "hard_link = %s", bool_to_string(conf->hard_link));
-	printer(s, conf->item_origins[find_conf("hard_link")->number], context);
-
-	reformat(&s, "hash_dir = %s", bool_to_string(conf->hash_dir));
-	printer(s, conf->item_origins[find_conf("hash_dir")->number], context);
-
-	reformat(&s, "ignore_headers_in_manifest = %s",
-	         conf->ignore_headers_in_manifest);
-	printer(s,
-	        conf->item_origins[find_conf("ignore_headers_in_manifest")->number],
-	        context);
-
-	reformat(&s, "keep_comments_cpp = %s",
-	         bool_to_string(conf->keep_comments_cpp));
-	printer(s, conf->item_origins[find_conf(
-	                                "keep_comments_cpp")->number], context);
-
-	reformat(&s, "limit_multiple = %.1f", (double)conf->limit_multiple);
-	printer(s, conf->item_origins[find_conf("limit_multiple")->number], context);
-
-	reformat(&s, "log_file = %s", conf->log_file);
-	printer(s, conf->item_origins[find_conf("log_file")->number], context);
-
-	reformat(&s, "max_files = %u", conf->max_files);
-	printer(s, conf->item_origins[find_conf("max_files")->number], context);
-
-	char *s2 = format_parsable_size_with_suffix(conf->max_size);
-	reformat(&s, "max_size = %s", s2);
-	printer(s, conf->item_origins[find_conf("max_size")->number], context);
-	free(s2);
-
-	reformat(&s, "path = %s", conf->path);
-	printer(s, conf->item_origins[find_conf("path")->number], context);
-
-	reformat(&s, "prefix_command = %s", conf->prefix_command);
-	printer(s, conf->item_origins[find_conf("prefix_command")->number], context);
-
-	reformat(&s, "prefix_command_cpp = %s", conf->prefix_command_cpp);
-	printer(s, conf->item_origins[find_conf(
-	                                "prefix_command_cpp")->number], context);
-
-	reformat(&s, "read_only = %s", bool_to_string(conf->read_only));
-	printer(s, conf->item_origins[find_conf("read_only")->number], context);
-
-	reformat(&s, "read_only_direct = %s", bool_to_string(conf->read_only_direct));
-	printer(s, conf->item_origins[find_conf("read_only_direct")->number],
-	        context);
-
-	reformat(&s, "recache = %s", bool_to_string(conf->recache));
-	printer(s, conf->item_origins[find_conf("recache")->number], context);
-
-	reformat(&s, "run_second_cpp = %s", bool_to_string(conf->run_second_cpp));
-	printer(s, conf->item_origins[find_conf("run_second_cpp")->number], context);
-
-	reformat(&s, "sloppiness = ");
-	if (conf->sloppiness & SLOPPY_FILE_MACRO) {
-		reformat(&s, "%sfile_macro, ", s);
-	}
-	if (conf->sloppiness & SLOPPY_INCLUDE_FILE_MTIME) {
-		reformat(&s, "%sinclude_file_mtime, ", s);
-	}
-	if (conf->sloppiness & SLOPPY_INCLUDE_FILE_CTIME) {
-		reformat(&s, "%sinclude_file_ctime, ", s);
-	}
-	if (conf->sloppiness & SLOPPY_TIME_MACROS) {
-		reformat(&s, "%stime_macros, ", s);
-	}
-	if (conf->sloppiness & SLOPPY_PCH_DEFINES) {
-		reformat(&s, "%spch_defines, ", s);
-	}
-	if (conf->sloppiness & SLOPPY_FILE_STAT_MATCHES) {
-		reformat(&s, "%sfile_stat_matches, ", s);
-	}
-	if (conf->sloppiness & SLOPPY_NO_SYSTEM_HEADERS) {
-		reformat(&s, "%sno_system_headers, ", s);
-	}
-	if (conf->sloppiness) {
-		// Strip last ", ".
-		s[strlen(s) - 2] = '\0';
-	}
-	printer(s, conf->item_origins[find_conf("sloppiness")->number], context);
-
-	reformat(&s, "stats = %s", bool_to_string(conf->stats));
-	printer(s, conf->item_origins[find_conf("stats")->number], context);
-
-	reformat(&s, "temporary_dir = %s", conf->temporary_dir);
-	printer(s, conf->item_origins[find_conf("temporary_dir")->number], context);
-
-	if (conf->umask == UINT_MAX) {
-		reformat(&s, "umask = ");
-	} else {
-		reformat(&s, "umask = %03o", conf->umask);
-	}
-	printer(s, conf->item_origins[find_conf("umask")->number], context);
-
-	reformat(&s, "unify = %s", bool_to_string(conf->unify));
-	printer(s, conf->item_origins[find_conf("unify")->number], context);
-
-	free(s);
-	return true;
+	bool ok = true;
+	ok &= print_item(conf, "base_dir", printer, context);
+	ok &= print_item(conf, "cache_dir", printer, context);
+	ok &= print_item(conf, "cache_dir_levels", printer, context);
+	ok &= print_item(conf, "compiler", printer, context);
+	ok &= print_item(conf, "compiler_check", printer, context);
+	ok &= print_item(conf, "compression", printer, context);
+	ok &= print_item(conf, "compression_level", printer, context);
+	ok &= print_item(conf, "cpp_extension", printer, context);
+	ok &= print_item(conf, "debug", printer, context);
+	ok &= print_item(conf, "direct_mode", printer, context);
+	ok &= print_item(conf, "disable", printer, context);
+	ok &= print_item(conf, "extra_files_to_hash", printer, context);
+	ok &= print_item(conf, "hard_link", printer, context);
+	ok &= print_item(conf, "hash_dir", printer, context);
+	ok &= print_item(conf, "ignore_headers_in_manifest", printer, context);
+	ok &= print_item(conf, "keep_comments_cpp", printer, context);
+	ok &= print_item(conf, "limit_multiple", printer, context);
+	ok &= print_item(conf, "log_file", printer, context);
+	ok &= print_item(conf, "max_files", printer, context);
+	ok &= print_item(conf, "max_size", printer, context);
+	ok &= print_item(conf, "path", printer, context);
+	ok &= print_item(conf, "pch_external_checksum", printer, context);
+	ok &= print_item(conf, "prefix_command", printer, context);
+	ok &= print_item(conf, "prefix_command_cpp", printer, context);
+	ok &= print_item(conf, "read_only", printer, context);
+	ok &= print_item(conf, "read_only_direct", printer, context);
+	ok &= print_item(conf, "recache", printer, context);
+	ok &= print_item(conf, "run_second_cpp", printer, context);
+	ok &= print_item(conf, "sloppiness", printer, context);
+	ok &= print_item(conf, "stats", printer, context);
+	ok &= print_item(conf, "temporary_dir", printer, context);
+	ok &= print_item(conf, "umask", printer, context);
+	ok &= print_item(conf, "unify", printer, context);
+	return ok;
 }
